"""
The `brent.datasets` module contains functions that will generate
datasets. Typically these datasets are meant for educational purposes.

Don't forget to see the docstring for a reference to the origin of the
dataset.
"""

import itertools as it
from functools import reduce

import pandas as pd
import numpy as np


def simple_study_dataset():
    """
    Creates a simple dataset that simulates the probability of passing
    a course given if the test for the course was hard and weather or
    not a student studied. This is a fake dataset.
    """
    grade = ["good", "pass", "fail"]
    study = ["lots", "little", "no"]
    hard = ["yes", "no"]
    probs = [0.6, 0.3, 0.1,
             0.9, 0.09, 0.01,
             0.3, 0.3, 0.4,
             0.6, 0.2, 0.2,
             0.1, 0.3, 0.6,
             0.1, 0.5, 0.4]
    zipprod = zip(it.product(study, hard, grade), probs)
    tween = reduce(lambda a, b: a + b, [[c] * int(p * 100) for c, p in zipprod])
    return pd.DataFrame(tween, columns=["study", "hard", "grade"])


def simple_sickness():
    """
    Inspiration of the data: https://www.cdc.gov/flu/about/qa/coldflu.htm.
           cold  flu  none astma
    sneeze sure   na  sure   bit
    cough   bit  yep  sure   yep
    weak   sure  yep    na   bit
    nose    yep sure  sure   bit
    chills   na  yep    na   bit
    tired  sure  yes  sure   yep
    """
    raise NotImplementedError("not yet implemented")


def asian_cancer_dataset():
    """
    This function reads in the asian cancer dataset. It is a dataset
    that has orignally also been featured in the `bnlearn` package. It
    is simulated data but has been used historically in benchmarks.

    The original description to the dataset can be found here:
    http://www.bnlearn.com/bnrepository/discrete-small.html#cancer

    ## Reference

    S. Lauritzen, D. Spiegelhalter. Local Computation with Probabilities on
    Graphical Structures and their Application to Expert Systems (with discussion).
    Journal of the Royal Statistical Society: Series B (Statistical Methodology),
    50(2):157-224, 1988.
    """
    return pd.read_csv("http://www.ccd.pitt.edu/wiki/images/ASIA10k.csv")


def alarm_dataset():
    """
    This function reads in the alarm dataset which has historically been
    used to describe alarms for monitoring stations for anesthesia.
    It is a dataset that has orignally also been featured in the `bnlearn` package.
    It is simulated data but has been used historically in benchmarks.

    The original description to the dataset can be found here:
    http://www.bnlearn.com/bnrepository/discrete-medium.html#alarm

    ## Reference:
    I. A. Beinlich, H. J. Suermondt, R. M. Chavez, and G. F. Cooper.
    The ALARM Monitoring System: A Case Study with Two Probabilistic Inference
    Techniques for Belief Networks. In Proceedings of the 2nd European Conference
    on Artificial Intelligence in Medicine, pages 247-256. Springer-Verlag, 1989.
    """
    return pd.read_csv("http://www.ccd.pitt.edu/wiki/images/ALARM10k.csv")


def blue_baby_dataset():
    """
    This function reads in the blue baby dataset. It involves a
    network for diagnosing congenital heart disease in a new born "blue baby".
    It is a dataset that has orignally also been featured in the `bnlearn` package.
    It is simulated data but has been used historically in benchmarks.

    The original description to the dataset can be found here:
    http://www.bnlearn.com/bnrepository/discrete-medium.html#child

    ## Reference
    D. J. Spiegelhalter, R. G. Coewll (1992). Learning in probabilistic expert systems.
    In Bayesian Statistics 4 (J. M. Bernardo, J. 0. Berger, A. P. Dawid and A. F. M.
    Smith, eds.) 447-466. Clarendon Press, Oxford.
    """
    return pd.read_csv("http://www.ccd.pitt.edu/wiki/images/CHILD10k.csv")


def earthquake_dataset():
    raise NotImplementedError("not yet implemented")


def generate_risk_dataset(attackers=3, defenders=2, battle_size=2):
    """
    This dataset generalises a scenario in the RISK board game. In this game
    typically three armies attack and two defend. The highest scoring attacker
    (based on a dice roll) is matched with the highest scoring defender, the second
    highest attacker goes with the second highest defender and so on. The defending
    party has the advantage so the attacker needs to roll higher than the defender
    in order to win. The `losses` column in the dataframe corresponds to the losses
    that the attacker incurs after the battle.

    This dataset is used in the corresponding `brent.examples.generate_risk_dag`.

    ## Input

    - **num_attackers**: The number of dice rolled by the attacker (default: 3)
    - **num_defenders**: The number of dice rolled by the defender (default: 2)
    - **num_attackers**: The number of armies that take part in the battle (default: 2)

    ## Output

    A dataframe with dice rolls, scores for participating armies and the loss outcome.
    """
    if min(attackers, defenders) < battle_size:
        raise ValueError(
            f"We demand min(num_attackers={attackers}, num_defenders={defenders}) >= battle_size={battle_size}.")
    attack_names = [f"a{i}" for i in range(1, attackers + 1)]
    defend_names = [f"d{i}" for i in range(1, defenders + 1)]
    combinations = it.product(*[range(1, 6+1) for i in range(attackers + defenders)])
    df = pd.DataFrame(combinations, columns=attack_names + defend_names)
    for side in ['a', 'd']:
        for b in range(1, battle_size + 1):
            names = attack_names if side == 'a' else defend_names
            df[f'best_{side}{b}'] = np.sort(df[names].values)[:, -b]

    best_attackers = df[[c for c in df.columns if 'best' in c and 'a' in c]].values
    best_defenders = df[[c for c in df.columns if 'best' in c and 'd' in c]].values

    df['losses'] = (best_attackers > best_defenders).sum(axis=1)
    return df



"""
The `brent.graph` module contains the DAG object. This is the
main object that you'll talk to when constructing a casual graph.
"""

import logging

import numpy as np
import pandas as pd
import networkx as nx
from graphviz import Digraph

from brent.common import normalise, window, is_path_blocked


class DAG:
    """
    A Directed-Acyclic-Graph (DAG) object describes a graphical model of a dataset.
    This object is generated from a pandas dataframe. Every column in the
    dataframe will result in a node/variable in the DAG object.

    ```
    from brent import DAG
    from brent.common import make_fake_df
    # let's start with a new dataset
    df = make_fake_df(4)
    dag = DAG(df).add_edge("a", "b").add_edge("b", "c").add_edge("c","d")
    ```
    """
    def __init__(self, dataframe: pd.DataFrame):
        """
        Create a new DAG from a dataframe.

        Inputs:

        - **dataframe**: pandas object that contains all variables

        Example:

        ```
        from brent import DAG
        from brent.common import make_fake_df
        # let's start with a new dataset
        df = make_fake_df(4)
        dag = DAG(df).add_edge("a", "b").add_edge("b", "c").add_edge("c","d")
        ```
        """
        self.df = dataframe
        self.graph = nx.DiGraph()
        for node in self.df.columns:
            self.graph.add_node(node)

    @property
    def undirected(self):
        """
        Fetch the `undirected` variant of the NetworkX graph. This can be
        useful when trying to determine all paths between two nodes.
        """
        return self.graph.to_undirected()

    @property
    def origin_nodes(self):
        """These nodes are nodes that do not have any edges going in."""
        return tuple(x for x in self.graph.nodes() if self.graph.in_degree(x) == 0)

    @property
    def marginal_table(self):
        """
        The marginal table is a table with all possible values and associated probability.
        """
        nodes = list(self.graph.nodes)
        logging.debug(f"about to calculate marginal table with nodes {nodes}")
        logging.debug(f"updating table for node {nodes[-1]}")
        marginal = self.calc_node_table(nodes.pop())
        logging.debug(f"current node table:\n{marginal}")
        logging.debug(f"current marginal table:\n{marginal}")
        for node in nodes:
            logging.debug(f"updating table for node {node}")
            logging.debug(f"current node table:\n{self.calc_node_table(node)}")
            marginal = self.merge_probs(marginal, self.calc_node_table(node))
            logging.debug(f"current marginal table:\n{marginal}")
        return marginal

    @property
    def nodes(self):
        """The nodes of the graph."""
        return list(self.graph.nodes)

    @property
    def edges(self):
        """The edges of the graph."""
        return list(self.graph.edges)

    def copy(self):
        """Returns a copy of the current DAG."""
        new_dag = DAG(self.df)
        new_dag.graph = self.graph
        return new_dag

    def edge_direction(self, node_a, node_b):
        """Determines the `<-` vs. `->` direction of an edge between two nodes."""
        if node_b in self.parents(node_a):
            return "<-"
        if node_a in self.parents(node_b):
            return "->"
        raise ValueError(f"node '{node_a}' is not connected to '{node_b}'")

    def undirected_paths(self, node_a, node_b):
        """
        Returns a list of all the paths that are between `node_a` and `node_b`.
        These paths do not take the direction into account and will turn the
        directed graph into an undirected one.
        """
        return list(nx.all_simple_paths(self.undirected, node_a, node_b))

    def directed_paths(self, node_a, node_b):
        output_paths = []
        for path in self.undirected_paths(node_a, node_b):
            windowed = list(window(path))
            new_path = []
            for n1, n2 in windowed:
                new_path.append(n1)
                new_path.append(self.edge_direction(n1, n2))
            new_path.append(n2)
            output_paths.append(new_path)
            logging.debug(f"found directed path: {' '.join(new_path)}")
        return output_paths

    def active_paths(self, node_a, node_b, z=list()):
        """
        Returns a list of all the paths that can influence probability between
        `node_a` and `node_b`. You can also supply a collection of nodes `z` that
        are given. These nodes might block (or enable) a path to be active. If there
        are no active paths between two nodes then this means that the two nodes
        are (conditionally if `z` is given) independant.
        """
        undirected_paths = self.directed_paths(node_a, node_b)
        active_paths = []
        for path in undirected_paths:
            z_path = [f"given({_})" if _ in z else _ for _ in path]
            logging.debug(f"now checking {' '.join(z_path)}")
            if not is_path_blocked(z_path):
                active_paths.append(z_path)
        return active_paths

    def calc_node_table(self, name):
        """
        Calculates probability table for a given node.

        Suppose we have graph `A -> B -> C`: `.calc_node_table("b")`
        call will calculate `P(B|A)` in the `probs` column of the result.

        ## Input

        - **name**: Name of a node/variable in the graph
        """
        parents = self.parents(name)
        tbl = self.df.copy()
        logging.debug(f"creating node table node={name} parents={parents}")

        def calculate_parents_size(dataf, groups=[]):
            return (dataf
                    .assign(count=1)
                    .groupby(groups)
                    .transform(np.sum)['count'])

        if len(parents) == 0:
            tbl = tbl.assign(parent_size=lambda d: d.shape[0])
        else:
            tbl = tbl.assign(parent_size=lambda d: calculate_parents_size(d, list(parents)))
        return (tbl
                .assign(node_size=lambda d: calculate_parents_size(d, list(parents) + [name]))
                .assign(prob=lambda d: d['node_size'] / d['parent_size'])
                [list(parents) + [name] + ["prob"]]
                .drop_duplicates()
                .reset_index(drop=True))

    def merge_probs(self, this_df, that_df):
        """
        Merges two probability dataframes while checking if nodes
        are connected in the graph. If `this_df` denotes `p(C|A,B)`
        in table form and `that_df` denotes `p(B|A)` in table form
        then the output of this function will denote `p(C,B|A)` in
        table form. Note that this method is mainly meant for internal usage.

        ## Input

        - **this_df**: Name of a node/variable in the graph, say `p(C|A,B)`
        - **that_df**: Name of a node/variable in the graph, say `p(B|A)`

        ## Output

        A dataframe that merges the two former tables, say `p(C,B|A)`.
        """
        common_cols = list(set(this_df.columns)
                           .intersection(set(that_df.columns))
                           .difference({"prob"}))
        if len(common_cols) == 0:
            columns = set(that_df.columns).difference({"prob"})
            loose_tables = []
            for c in columns:
                for value in self.calc_node_table(c)[c].values:
                    loose_tables.append(this_df.assign(**{c: value}))
            join_able = pd.concat(loose_tables)
            return self.merge_probs(join_able, that_df)
        return (this_df
                .set_index(common_cols)
                .join(that_df.set_index(common_cols), lsuffix="1", rsuffix="2")
                .assign(prob=lambda x: normalise(x.prob1 * x.prob2))
                .drop("prob1", axis=1)
                .drop("prob2", axis=1)
                .reset_index())

    def add_edge(self, source, sink) -> 'DAG':
        """
        Adds an edge to the graph.

        ## Input

        - **source**: Name of a node in the graph
        - **sink**: Name of a node in the graph

        ## Example

        ```
        from brent import DAG
        from brent.common import make_fake_df

        (DAG(dataframe=make_fake_df(4))
            .add_edge("a", "b")
            .add_edge("b", "c")
            .add_edge("c", "d"))
        ```
        """
        if source not in self.df.columns:
            raise ValueError(f"cause {source} not in dataframe")
        if sink not in self.df.columns:
            raise ValueError(f"effect {sink} not in dataframe")
        new_graph = self.graph.copy()
        new_graph.add_edge(source, sink)
        if not nx.is_directed_acyclic_graph(new_graph):
            raise ValueError(f"edge {source} -> {sink} causes DAG to get cycle")
        self.graph.add_edge(source, sink)
        logging.debug(f"created connection {source} -> {sink}")
        return self

    def children(self, node):
        """
        Return the children of a node.

        ## Input

        - **node**: Name of a node

        ## Example

        ```
        from brent import DAG
        from brent.common import make_fake_df

        dag = (DAG(dataframe=make_fake_df(4))
            .add_edge("a", "b")
            .add_edge("b", "c")
            .add_edge("c", "d"))

        dag.children("b") #outputs "c"
        dag.children("c") #outputs "d"
        ```
        """
        return set(self.graph.successors(node))

    def parents(self, node):
        """
        Return the parents of a node.

        ## Input

        - **node**: Name of a node

        ## Example

        ```
        from brent import DAG
        from brent.common import make_fake_df

        dag = (DAG(dataframe=make_fake_df(4))
            .add_edge("a", "b")
            .add_edge("b", "c")
            .add_edge("c", "d"))

        dag.children("b") #outputs "a"
        dag.children("c") #outputs "b"
        ```
        """
        return set(self.graph.predecessors(node))

    def connections(self, node):
        """
        Return all nodes connected to the one passed in.

        ## Input

        - **node**: Name of a node

        ## Example

        ```
        from brent import DAG
        from brent.common import make_fake_df

        dag = (DAG(dataframe=make_fake_df(4))
            .add_edge("a", "b")
            .add_edge("b", "c")
            .add_edge("c", "d"))

        dag.children("b") #outputs ["a","c"]
        dag.children("c") #outputs ["b","d"]
        ```
        """
        return set(list(self.children(node)) + list(self.parents(node)))

    def independences(self):
        """
        **NOT IMPLEMENTED YET**
        """
        # https://www.slideshare.net/duytungpham18/lecture-1-graphical-models
        pass

    def plot(self):
        """A pretty plotting function."""
        d = Digraph()
        d.attr(rankdir='LR')
        d.attr('node', shape='circle')
        for n in self.graph.nodes:
            d.node(n)
        for n1, n2 in self.graph.edges:
            d.edge(n1, n2)
        return d

    def nx_plot(self, **kwargs):
        """
        A customizable plotting function. The function comes from `networkx`.
        It merely wraps around the `nx.draw` method, documentation of this project
        can be found [here](https://networkx.github.io/documentation/stable/index.html).
        """
        nx.draw(self.graph, node_size=500, with_labels=True, node_color="white", **kwargs)



from brent import DAG
from brent.datasets import generate_risk_dataset


def generate_risk_dag(attackers=3, defenders=2, battle_size=2):
    """
    This DAG generalises a scenario in the RISK board game. In this game
    typically three armies attack and two defend. The highest scoring attacker
    (based on a dice roll) is matched with the highest scoring defender, the second
    highest attacker goes with the second highest defender and so on. The defending
    party has the advantage so the attacker needs to roll higher than the defender
    in order to win. The `losses` node in the graph corresponds to the losses
    that the attacker incurs after the battle.

    This dag is used in the corresponding `brent.datasets.generate_risk_dataset`.

    ## Input

    - **num_attackers**: The number of dice rolled by the attacker (default: 3)
    - **num_defenders**: The number of dice rolled by the defender (default: 2)
    - **num_attackers**: The number of armies that take part in the battle (default: 2)

    ## Output

    A DAG object with correct arcs, ready for queries.
    """
    if min(attackers, defenders) < battle_size:
        raise ValueError(
            f"We demand min(num_attackers={attackers}, num_defenders={defenders}) >= battle_size={battle_size}.")
    attack_names = [f"a{i}" for i in range(1, attackers + 1)]
    defend_names = [f"d{i}" for i in range(1, defenders + 1)]
    dag = DAG(dataframe=generate_risk_dataset(attackers, defenders, battle_size))
    for side in ['a', 'd']:
        for b in range(1, battle_size + 1):
            names = attack_names if side == 'a' else defend_names
            for name in names:
                dag.add_edge(name, f"best_{side}{b}")
    for n in [_ for _ in dag.nodes if 'best' in _]:
        dag.add_edge(n, 'losses')
    return dag



"""
Brent
-----

![](https://github.com/koaning/brent/blob/master/images/dag1.png?raw=true)

Brent is a playful tool to help understand bayesian graphical
modelling as well as do-calculus. It offers support for discrete
values and comes with great plotting tools. Under the hood is merely
uses [pandas](http://pandas.pydata.org/pandas-docs/stable/reference/index.html)
and [networkx](https://networkx.github.io/documentation/stable/index.html).

## Example 1

The basics of the tool can be demonstrated via the following codeblock.

```python
from brent import DAG, Query
from brent.common import make_fake_df

# let's start with a new dataset and visualise it's graph
df = make_fake_df(4)
dag = DAG(df).add_edge("a", "b").add_edge("b", "c").add_edge("c","d")
dag.plot()
```

![](https://github.com/koaning/brent/blob/master/images/simple-graph.png?raw=true)

Such a graph is nice and we can get properties from it, these can be
found in the documentation page. The main usecase for it though is
to use

```python
# we can build a query dynamically
q = Query(dag).given(a=1).do(d=1)
q.plot()
```

![](https://github.com/koaning/brent/blob/master/images/simple-query.png?raw=true)

The nodes that are given have double circles around them while the
nodes that are affected by a `do()`-operation have greyed out arcs
going in. This is because these are usually removed for inference.

If you're interested in inferring probabilities, you can get the
updated probabilities by calling the inference method on the object.

```python
# we can also see updated probabilities
q.infer()
```

## Example 2

We'll still be using fake data for this example, but we're going to
create a more complex graph.

```
from brent import DAG, Query
from brent.common import make_fake_df

dag = (DAG(make_fake_df(7))
       .add_edge("e", "a")
       .add_edge("e", "d")
       .add_edge("a", "d")
       .add_edge("b", "d")
       .add_edge("a", "b")
       .add_edge("a", "c")
       .add_edge("b", "c")
       .add_edge("c", "f")
       .add_edge("g", "f"))

dag.plot()
```

![](https://github.com/koaning/brent/blob/master/images/complex-graph.png?raw=true)

This graph will be more involved when we create a query on it.

```
q = Query(dag).given(d=1).do(a=0, c=1)
q.plot()
```

But we can still make a pretty plot appear.

![](https://github.com/koaning/brent/blob/master/images/complex-query.png?raw=true)

Inference is still done in the same way. Note that you can
get marginal probabilities in json format but you can also
get pass `give_table=True` if you prefer to get a full probability
table. This will give a pandas dataframe.

```python
# we can also see updated probabilities
q.infer()
q.infer(give_table=True)
```

"""

from .query import Query
from .graph import DAG



"""
The `brent.query` module contains the `Query` object. This is the
main object that you'll use to describe complex queries against
casual graphs.
"""

import logging

import numpy as np
from graphviz import Digraph

from brent.graph import DAG
from brent.common import normalise


class Query:
    def __init__(self, dag: DAG, given=None, do=None, counterfact=None):
        """
        A Query object describes a query that will be run on a DAG object.

        Inputs:

        - **dag**: DAG object that contains all edges.
        - **given**: Dictionary of key-value pairs of given items.
        - **do**: Dictionary of key-value pairs of do operated items.
        - **counterfact**: Dictionary of key-value pairs of counterfact operated items.

        Example:

        ```
        from brent import DAG, Query
        from brent.common import make_fake_df
        # let's start with a new dataset
        df = make_fake_df(4)
        dag = DAG(df).add_edge("a", "b").add_edge("b", "c").add_edge("c","d")
        # we can build a query dynamically
        q1 = Query().given(a=1).do(d=1)
        # alternatively we can build one from start
        q2 = Query(given={'a':1}, do={'d':1})
        ```
        """
        self.dag = dag
        if not given:
            given = dict()
        if not do:
            do = dict()
        if not counterfact:
            counterfact = dict()
        self.given_dict = given
        self.do_dict = do
        self.counterfact = counterfact

    def inference_dag(self):
        """
        This is a DAG created from the original but has been altered
        to accomodate `do-calculus`.
        """
        infer_dag = DAG(self.dag.df.copy())
        logging.debug(f"constructing copy of original DAG nodes: {infer_dag.nodes}")
        for n1, n2 in self.dag.edges:
            if n2 not in self.do_dict.keys():
                infer_dag.add_edge(n1, n2)
            else:
                logging.debug(f"edge {n1} -> {n2} ignored because of do operator")
        logging.debug(f"original DAG copied")
        return infer_dag

    def _check_query_input(self, **kwargs):
        for key, value in kwargs.items():
            logging.debug(f"checking key {key}={value}")
            if key not in self.dag.nodes:
                raise ValueError(f"node {key} does not exist in original dag")
            if value not in self.dag.df[key].values:
                raise ValueError(f"value {value} does not occur for node {key}")
            if key in {**self.given_dict, **self.do_dict}.keys():
                raise ValueError(f"{key} is already used in this query")

    def given(self, **kwargs):
        """
        Add items to the query that are `given`.

        ## Inputs

        - **kwargs**: key-value pairs of given items.
        """
        self._check_query_input(**kwargs)
        return Query(dag=self.dag, given={**self.given_dict, **kwargs}, do=self.do_dict, counterfact=self.counterfact)

    def do(self, **kwargs):
        """
        Add items to the query that are enforced via `do` operation.

        ## Inputs

        - **kwargs**: key-value pairs of do items.
        """
        self._check_query_input(**kwargs)
        return Query(dag=self.dag, given=self.given_dict, do={**self.do_dict, **kwargs}, counterfact=self.counterfact)

    def counterfact(self, **kwargs):
        """
        Add items to the query that are observed but will be counterfacted.

        ## Inputs

        - **kwargs**: key-value pairs of observations to be counterfacted.
        """
        self._check_query_input(**kwargs)
        return Query(dag=self.dag, given=self.given_dict, do=self.do_dict, counterfact={**self.counterfact, **kwargs})

    def plot(self, emphesize_do=True):
        """
        A pretty plotting function. Given nodes have double circles.
        Nodes with `do` operations on them will have in-going arcs grayed.
        """
        givens = self.given_dict.keys()
        dos = self.do_dict.keys()
        d = Digraph()
        d.attr(rankdir='LR')
        d.attr('node', shape='circle')
        for idx, n in enumerate(self.dag.graph.nodes):
            if (n in givens) or (n in dos):
                d.node(n, shape='doublecircle')
            if n in dos:
                if emphesize_do:
                    d.node(" " * idx, shape="none")
                    d.edge(" " * idx, n)
            else:
                d.node(n)
        for n1, n2 in self.dag.graph.edges:
            if ((n1 in dos) or (n1 in givens)) and ((n2 in dos) or (n2 in givens)):
                d.edge(n1, n2, color="lightgray")
            elif n2 in dos:
                d.edge(n1, n2, color="lightgray", style="dashed")
            else:
                d.edge(n1, n2)
        return d

    def infer_counterfact(self):
        """
        If counterfact is in the query we will need to return.

        The counterfacted variable is first assumed to be "given". We apply inference
        on the entire graph as a first step. This gives us probabilities for all values.
        This is the new graph definition.


        """
        pass

    def infer(self, give_table=False):
        """
        Run the inference on the graph given the current query.

        ## Inputs

        - **give_table**: Instead of calculating marginal probabilities and
        returning a dictionary, return a pandas table instead. Defaults to `False`.
        """
        logging.debug(f"about to make an inference")
        infer_dag = self.inference_dag()
        for node in infer_dag.nodes:
            logging.debug(f"confirming parents({node}) = {infer_dag.parents(node)}")
        marginal_table = infer_dag.marginal_table
        for k, v in {**self.do_dict, **self.given_dict}.items():
            logging.debug(f"processing {k}={v}")
            marginal_table = marginal_table.loc[lambda d: d[k] == v]
        tbl = marginal_table.assign(prob=lambda d: normalise(d.prob))
        if give_table:
            return tbl
        output = {}
        for c in tbl.columns:
            if c != "prob":
                output[c] = tbl.groupby(c)['prob'].sum().to_dict()
        return output

    def sample(self, n_samples=1):
        """
        Sample data from the current query.

        ## Inputs

        - **n_samples**: the number of samples to get

        ## Output

        `pandas.DataFrame` with new samples

        ## Example

        Example:

        ```
        from brent import DAG, Query
        from brent.common import make_fake_df
        # let's start with a new dataset
        df = make_fake_df(4)
        dag = DAG(df).add_edge("a", "b").add_edge("b", "c").add_edge("c","d")
        # we can build a query dynamically
        q1 = Query().given(a=1).do(d=1)
        q1.sample(100)
        ```
        """
        table = self.infer(give_table=True)
        idx = np.random.choice(table.index, p=table.prob, replace=True, size=n_samples)
        return table.loc[idx].reset_index(drop=True).drop(columns=['prob'])



import os
from setuptools import setup, find_packages


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


base_packages = ["numpy>=1.15.4", "scipy>=1.2.0", "scikit-learn>=0.20.2",
                 "pandas>=0.23.4", "matplotlib>=3.0.2", "networkx>=2.2",
                 "graphviz>=0.10.1"]

setup(
    name="brent",
    version="0.2.3",
    packages=find_packages(exclude=['data', 'notebooks']),
    long_description=read('readme.md'),
    keywords=['causal', 'bayesian', 'graphical', 'model', 'inference'],
    python_requires='>=3.6',
    install_requires=base_packages,
    extras_require={
        "dev": ["flake8>=3.6.0", "pytest>=3.3.1", "pdoc3>=0.5.2",
                "nbval>=0.9.1", "plotnine>=0.5.1", "twine>=1.13.0"]
    },
    classifiers=['Intended Audience :: Developers',
                 'Intended Audience :: Science/Research',
                 'Programming Language :: Python :: 3.6',
                 'Programming Language :: Python :: 3.7',
                 'Development Status :: 3 - Alpha',
                 'License :: OSI Approved :: MIT License',
                 'Topic :: Scientific/Engineering',
                 'Topic :: Scientific/Engineering :: Artificial Intelligence']
)



"""
The `brent.common` module contains common functions that can be
used while working with dataframes and brent graphs. They are also
used internally by the library.
"""
import logging
from itertools import islice

import numpy as np
import pandas as pd


def make_fake_df(nodes=6, rows=100, values=2, seed=42):
    """
    Creates a fake and random dataframe that can be used for demos.

    ## Inputs:

    - **nodes**: the number of nodes/variables to be generated
    - **rows**: the number of rows of fake data to generate
    - **values**: the different values that the variables can take
    - **seed**: the seed value for the random numbers to be generated

    ## Example

    ```
    from brent.common import make_fake_df
    # let's start with a new dataset
    df = make_fake_df(nodes=4, rows=1000, values=4, seed=41)
    ```
    """
    letters = 'abcdefghijklmnopqrstuvwxyz'
    np.random.seed(seed)
    if nodes > 26:
        raise ValueError('only allow 26 letters in the alfabet')
    return pd.DataFrame({k: np.random.randint(0, values, rows) for k in letters[:nodes]})


def normalise(x):
    """
    Simply normalises a numpy-like array or pandas-series.

    ## Inputs
    - **x**: a numpy array of pandas series

    ## Example

    ```
    import numpy as np
    from brent.common import normalise
    normalise(np.array([1,2,3,4]))
    ```
    """
    if isinstance(x, list):
        x = np.array(x)
    return x / x.sum()


def quantize_column(column, parts=4):
    """
    Turns a continous dataset into a discrete one by splitting
    it into quantiles.

    ## Inputs
    - **column**: a numpy array of pandas series
    - **parts**: the number of parts to split the data into

    ## Example

    ```
    import numpy as np
    from brent.common import quantize_column
    quantize_column(np.array([1,2,3,4]), parts=2)
    ```
    """
    return pd.cut(column, parts, labels=range(1, parts+1))


def window(seq, n=2):
    """
    Calculates a moving window over an iterable.

    ## Inputs
    - **seq**: an iterable sequence
    - **n**: the size of the window, typically this is equal to 2

    ## Example

    ```
    from brent.common import window

    list(window([1,2,3,4), n=2))
    ```
    """
    it = iter(seq)
    result = tuple(islice(it, n))
    if len(result) == n:
        yield result
    for elem in it:
        result = result[1:] + (elem,)
        yield result


def check_node_blocking(arrow_before, arrow_after, name):
    """
    Checks if a node is a blocking node. As a side effects logs can be
    written if you listen to `logging.debug`.

    ## Inputs
    - **arrow_before**: the direction of the arrow before the node
    - **arrow_after**: the direction of the arrow after the node
    - **name**: name of the node in question, if the string "given" is
                in it we will assume that it is given.

    ## Output
    Return `True`/`False`.

    ## Example

    ```
    from brent.common import check_node_blocking

    check_node_blocking("->", "->", "given_a") # True
    check_node_blocking("->", "->", "a") # False
    ```
    """
    given = "given" in name
    if (arrow_before == '<-') and (arrow_after == '->'):
        blocking = True if given else False
        logging.debug(f"checking: ... {arrow_before} {name} {arrow_after} ... type: `split` blocking: {blocking}")
    elif (arrow_before == '->') and (arrow_after == '<-'):
        blocking = False if given else True
        logging.debug(f"checking: ... {arrow_before} {name} {arrow_after} ... type: `collider` blocking: {blocking}")
    elif arrow_before == arrow_after:
        blocking = True if given else False
        logging.debug(f"checking: ... {arrow_before} {name} {arrow_after} ... type: `chain` blocking: {blocking}")
    else:
        raise ValueError(f"check arrow_before/arrow_after now:{arrow_before}, {arrow_after}")
    return blocking


def is_path_blocked(path_list):
    """
    Given a list of nodes and arcs, this function checks if the path is
    probabilistically blocked. We check if the path is blocked between
    the first and last element.

    ## Inputs
    - **path_list**: iterable of node_names (which might have the `given_` prefix
                     to indicate that it is given) which are alternated by arrows
                     ("->" or "<-") indicating the direction of the arcs on the path

    ## Output

    Return `True`/`False`.

    ## Example

    ```
    from brent.common import is_path_blocked

    check_node_blocking(["a", "->", "b", "->", "c") # False
    check_node_blocking(["a", "->", "given_b", "->", "c") # True
    ```
    """
    for idx, name in enumerate(path_list):
        if idx in [0, len(path_list) - 1]:
            pass
        elif name in ['<-', '->']:
            pass
        else:
            arrow_before = path_list[idx - 1]
            arrow_after = path_list[idx + 1]
            blocking = check_node_blocking(arrow_before, arrow_after, name)
            if blocking:
                logging.info("found blocking node, can skip path")
                return True
    return False


def join_independent(this_df: pd.DataFrame, that_df: pd.DataFrame) -> pd.DataFrame:
    """
    Merges two probability dataframes assuming independent nodes

    ## Example:

    ```
    >>> this_df = pd.DataFrame({'A': ['true', 'false'], 'prob': [0.5, 0.5]})
    >>> that_df = pd.DataFrame({'B': ['true', 'false'], 'prob': [0.5, 0.5]})
    >>> join_independent(this_df, that_df) # doctest: +NORMALIZE_WHITESPACE
           A      B  prob
    0   true   true  0.25
    1   true  false  0.25
    2  false   true  0.25
    3  false  false  0.25
    ```
    """
    if 'prob' not in this_df.columns:
        raise ValueError('this_df should contain a `prob` column containing probabilities')
    if 'prob' not in that_df.columns:
        raise ValueError('that_df should contain a `prob` column containing probabilities')

    return (this_df.assign(key=1)
            .merge(that_df.assign(key=1), on='key')
            .drop('key', 1)
            .assign(prob=lambda d: d.prob_x * d.prob_y)
            .drop(columns=['prob_x', 'prob_y'])
            )


def join_dependent(this_df: pd.DataFrame, that_df: pd.DataFrame) -> pd.DataFrame:
    """
    Merges two probability dataframes assuming dependencies between nodes by using that_df's indexes as conditionals.

    If `this_df` denotes `p(A)` in table form and `that_df` denotes `p(B|A)` in table form
    then the output of this function will denote `p(A, B)`.

    `that_df` should have two columns `B` and `prob` and its index should be set to `A`'s values.

    ## Example:

    ```
    >>> this_df = pd.DataFrame({'A': ['true', 'false'], 'prob': [0.5, 0.5]})
    >>> that_df = pd.DataFrame({
    ... 'A': ['true', 'false', 'true', 'false'],
    ... 'B': ['true', 'true', 'false', 'false'],
    ... 'prob': [0.3, 0.7, 0.8, 0.2]}).set_index('A')
    >>> join_dependent(this_df, that_df) # doctest: +NORMALIZE_WHITESPACE
           A      B  prob
    0   true   true  0.15
    0   true  false  0.40
    1  false   true  0.35
    1  false  false  0.10
    ```
    """
    if 'prob' not in this_df.columns:
        raise ValueError('this_df should contain a `prob` column containing probabilities')
    if 'prob' not in that_df.columns:
        raise ValueError('that_df should contain a `prob` column containing probabilities')

    missing_names = [name for name in that_df.index.names if name not in this_df.columns]
    if len(missing_names) > 0:
        raise ValueError('missing_names are set as indexes to `that_df` but are not present in `this_df`')

    if len(that_df.columns) > 2:
        raise ValueError('`that_df` has more than two columns, perhaps you forgot to set the variables that '
                         'are to be conditioned on as the index of the dataframe')

    return (this_df
            .merge(that_df, left_on=that_df.index.names, right_index=True)
            .assign(prob=lambda d: d.prob_x * d.prob_y)
            .drop(columns=['prob_x', 'prob_y'])
            .reset_index(drop=True)
            )



"""
The `sklearn` module contains objects that be used in scikit-learn pipelines.
In particulate it offers a classifier as well as an imputer.
"""

from brent.graph import DAG
from brent.query import Query

import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator, ClassifierMixin


class BrentClassifier(BaseEstimator, ClassifierMixin):
    """
    A classifier that allows you to define your own model via a DAG.
    """
    def __init__(self, dag: DAG, to_predict: str):
        """
        Construct an estimator based on a DAG. You need to specify the DAG as well as
        the column name that requires prediction.

        ## Inputs

        - **dag**: DAG object that describes the dag
        - **to_predict**: the column to predict

        ## Output

        A classifier that can be used in scikit-learn pipelines.
        """
        if to_predict not in dag.df.columns:
            raise ValueError(f"column {to_predict} not found in DAG {dag}")
        self.dag = dag
        self.to_predict = to_predict
        self.to_use = [_ for _ in self.dag.df.columns if _ != self.to_predict]
        self.query = None
        self.k = self.dag.df[to_predict].nunique()

    def _check_dataframe(self, X):
        for node in self.dag.nodes:
            if node not in X.columns:
                raise ValueError(f"column {node} not in dataframe but in DAG")

    def fit(self, X: pd.DataFrame, y):
        """
        Make the estimator "train". This is a bit verbose since the DAG object
        is already pretrained. We mainly check if the supplied dataframe given in `X`
        is consistent with the graph.

        ## Inputs

        - **X**: a dataframe to be used
        - **y**: ignored but required by the api

        ## Output

        A "trained" classifier that can be used in scikit-learn pipelines.
        """
        self._check_dataframe(X)
        return self

    def predict(self, X):
        """
        Predict the class.

        ## Inputs

        - **X**: a dataframe to be used

        ## Output

        A numpy array containing the predicted classes.
        """
        return np.argmax(self.predict_proba(X), axis=1)

    def predict_proba(self, X):
        """
        Predict the probabilities for all classes

        ## Inputs

        - **X**: a dataframe to be used

        ## Output

        A numpy array (num_rows, num_classes) containing the predicted classes.
        """
        self._check_dataframe(X)
        predictions = np.zeros((X.shape[0], self.k))
        for idx, row in X[self.to_use].reset_index(drop=True).iterrows():
            query = Query(dag=self.dag, given=row.to_dict())
            table = query.infer(give_table=True).sort_values(self.to_predict)[[self.to_predict, 'prob']]
            for i, r in table.iterrows():
                k = r[self.to_predict]
                predictions[idx, int(k)] = r['prob']
        return predictions



import re
import logging
from typing import Tuple, List, Iterator

import pandas as pd

from brent.common import join_independent, join_dependent

logger = logging.getLogger(__name__)


PARSER_PATTERNS = {
    'network': r'network (?P<networktype>\w+) {',
    'variable': r"(variable (?P<varname>\w+) {\s+type (?P<vartype>\w+) \[ \d+ \] \{ (?P<states>[^}]*)\}\;)",
    'uncond_probability': r"probability \( (?P<varname>\w+) \) \{\s+table (?P<probtable>[^;]*)",
    'cond_probability': r"probability \( (?P<varname>\w+) \| (?P<conditionals>[^)]*) \) {\s (?P<probs>[^\}]*)",
    'prob_table': r"^(\s+)? \((?P<keys>.*?)\) (?P<probabilities>[^;]*);",
}


def parse_network_type(bif: str) -> str:
    """Returns the network type of a bif string"""
    return re.match(PARSER_PATTERNS['network'], bif).group('networktype')


def parse_variables(bif: str) -> Iterator[Tuple[str, List[str]]]:
    """Returns a generator of `(varname, states)` tuples from a bif string"""
    for var in re.finditer(PARSER_PATTERNS['variable'], bif):
        varname = var.group('varname')
        vartype = var.group('vartype')
        states = var.group('states').strip().split(', ')

        logger.info(f'Found variable: {varname} with states {states}')
        if vartype != 'discrete':
            raise ValueError('Brent only supports discrete variables')

        yield varname, states


def parse_unconditional_probabilities(bif: str) -> Iterator[Tuple[str, List[float]]]:
    """Parses unconditional probabilities from a bif string"""
    for prob in re.finditer(PARSER_PATTERNS['uncond_probability'], bif):
        varname = prob.group('varname')
        probs = [float(p) for p in prob.group('probtable').split(', ')]

        logger.info(f'Found base probability for variable {varname}: {probs}')
        if 1 - sum(probs) > 0.01:
            logger.warning(f'Variable {varname} has probabilities not summing to 1 (sums to {sum(probs)})')

        yield varname, probs


def parse_conditional_probabilities(bif: str):
    """parses conditional probabilities from a bif string"""
    for cond_prob in re.finditer(PARSER_PATTERNS['cond_probability'], bif):
        varname, prob_table = cond_prob.group('varname'), cond_prob.group('probs')
        conditionals = cond_prob.group('conditionals').split(', ')
        logger.info(f'Found conditional probability: {varname} | {conditionals}')

        probabilities_lst = []

        for line_no, prob_entry in enumerate(re.finditer(PARSER_PATTERNS['prob_table'], prob_table, re.MULTILINE)):
            keys = prob_entry.group('keys').split(', ')
            probabilities = [float(p) for p in prob_entry.group('probabilities').split(', ')]
            if 1 - sum(probabilities) > 0.01:
                logger.warning(f'Variable {varname} | {conditionals} has probabilities not summing to 1 '
                               f'in line {line_no + 1} (sums to {sum(probabilities)})')

            logger.info(f'\t{keys} - {probabilities}')
            probabilities_lst.append((keys + probabilities))

        yield varname, conditionals, probabilities_lst


def bif(bif: str) -> Tuple[pd.DataFrame, List[Tuple[str, str]]]:
    """
    Parses a Bayesian Interchange Format (bif) string into a probability table and a set of links
    """
    probability_table = pd.DataFrame({'key': [1], 'prob': [1]})

    if parse_network_type(bif) != 'unknown':
        logger.warning("BIF Parser does not take into account known networks")

    variable_states = dict(parse_variables(bif))
    edges = []

    for varname, probs in parse_unconditional_probabilities(bif):
        var_df = pd.DataFrame({varname: variable_states[varname], 'prob': probs})
        probability_table = join_independent(probability_table, var_df)

    for varname, conditionals, probabilities in parse_conditional_probabilities(bif):
        cond_df = (pd.DataFrame(probabilities, columns=conditionals + variable_states[varname])
                   .melt(id_vars=conditionals, var_name=varname, value_name='prob')
                   .set_index(conditionals)
                   )
        probability_table = join_dependent(probability_table, cond_df)
        edges += [(conditional, varname) for conditional in conditionals]

    return probability_table, edges
